@@ -1,6 +1,11 @@
 #define PY_SSIZE_T_CLEAN
 #include <Python.h>
 
+/* "restrict" qualifier and "inline" function specifier were introduced in C99.
+ * See http://www.open-std.org/jtc1/sc22/wg14/www/docs/n1124.pdf
+ *
+ * Major compilers have alternatives of these keywords available even in C89.
+ */
 #if __STDC_VERSION__ < 199901L
 #ifdef __restrict
 #define restrict __restrict
@@ -14,36 +19,60 @@
 #endif /* __inline__ */
 #endif /* __STDC_VERSION__ < 199901L */
 
+/* Backport of the convenience macro introduced in Python 3.5
+ * See https://github.com/python/cpython/blob/v3.5.0/Include/pymacro.h#L19
+ */
 #ifndef Py_STRINGIFY
 #define ARC4_XSTRINGIFY(x) #x
 #define Py_STRINGIFY(x) ARC4_XSTRINGIFY(x)
-#endif
+#endif /* Py_STRINGIFY */
 
 struct arc4_state {
     unsigned char x, y, s[256];
 };
 
+static const struct arc4_state arc4_initial_state = {
+    0, /* x */
+    0, /* y */
+    {0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b,
+     0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
+     0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20, 0x21, 0x22, 0x23,
+     0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
+     0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b,
+     0x3c, 0x3d, 0x3e, 0x3f, 0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,
+     0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50, 0x51, 0x52, 0x53,
+     0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
+     0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x6b,
+     0x6c, 0x6d, 0x6e, 0x6f, 0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,
+     0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f, 0x80, 0x81, 0x82, 0x83,
+     0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
+     0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b,
+     0x9c, 0x9d, 0x9e, 0x9f, 0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
+     0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xb0, 0xb1, 0xb2, 0xb3,
+     0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
+     0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xcb,
+     0xcc, 0xcd, 0xce, 0xcf, 0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7,
+     0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf, 0xe0, 0xe1, 0xe2, 0xe3,
+     0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
+     0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa, 0xfb,
+     0xfc, 0xfd, 0xfe, 0xff} /* s */
+};
+
 static inline void
 arc4_init(struct arc4_state *state, const unsigned char *key,
           Py_ssize_t key_size)
 {
+    register unsigned char *s = state->s;
     register int i;
+    register unsigned char j = 0;
 
-    state->x = 0;
-    state->y = 0;
+    memcpy(state, &arc4_initial_state, sizeof(*state));
     for (i = 0; i < 256; i++) {
-        state->s[i] = (unsigned char)i;
-    }
-    {
-        register unsigned char j, k;
+        register unsigned char k = s[i];
 
-        j = 0;
-        for (i = 0; i < 256; i++) {
-            j += state->s[i] + key[i % key_size];
-            k = state->s[i];
-            state->s[i] = state->s[j];
-            state->s[j] = k;
-        }
+        j += k + key[i % key_size];
+        s[i] = s[j];
+        s[j] = k;
     }
 }
 
@@ -51,19 +80,17 @@ static inline void
 arc4_crypt(struct arc4_state *state, const unsigned char *restrict input,
            unsigned char *restrict output, Py_ssize_t size)
 {
-    register unsigned char x, y, *s;
-
-    x = state->x;
-    y = state->y;
-    s = state->s;
+    register unsigned char x = state->x;
+    register unsigned char y = state->y;
     {
+        register unsigned char *s = state->s;
         register unsigned char sx, sy;
         register Py_ssize_t i;
 
         for (i = 0; i < size; i++) {
             x++;
-            y += s[x];
             sx = s[x];
+            y += sx;
             sy = s[y];
             s[x] = sy;
             s[y] = sx;
@@ -102,7 +129,7 @@ arc4_ARC4_init(struct arc4_ARC4 *self, PyObject *args, PyObject *kwargs)
                      args->ob_type->tp_name);
         return -1;
     }
-#endif
+#endif /* PYPY_VERSION */
     if (key_size <= 0) {
         PyErr_Format(PyExc_ValueError, "invalid key length: %zd", key_size);
         return -1;
@@ -120,40 +147,6 @@ arc4_ARC4_dealloc(struct arc4_ARC4 *self)
     Py_TYPE(self)->tp_free((PyObject *)self);
 }
 
-static inline int
-byteslike_as_string_and_size(PyObject *obj, char **buffer, Py_ssize_t *size)
-{
-#if PY_MAJOR_VERSION >= 3
-    const char *utf8;
-#else
-    PyObject *ascii;
-#endif
-
-    if (PyBytes_Check(obj)) {
-        *buffer = PyBytes_AS_STRING(obj);
-        *size = PyBytes_GET_SIZE(obj);
-        return 0;
-    }
-    if (PyUnicode_Check(obj)) {
-#if PY_MAJOR_VERSION >= 3
-        utf8 = (char *)PyUnicode_AsUTF8AndSize(obj, size);
-        if (utf8 != NULL) {
-            *buffer = (char *)utf8;
-            return 0;
-        }
-#else
-        ascii = PyUnicode_AsASCIIString(obj);
-        if (ascii != NULL) {
-            *buffer = PyBytes_AS_STRING(ascii);
-            *size = PyBytes_GET_SIZE(ascii);
-            Py_DECREF(ascii);
-            return 0;
-        }
-#endif
-    }
-    return -1;
-}
-
 static PyObject *
 arc4_ARC4_crypt(struct arc4_ARC4 *self, PyObject *arg)
 {
@@ -165,14 +158,31 @@ arc4_ARC4_crypt(struct arc4_ARC4 *self, PyObject *arg)
     if (arg == NULL) {
         return NULL;
     }
-    if (byteslike_as_string_and_size(arg, (char **)&input, &size) == -1) {
-        if (PyErr_Occurred() == NULL) {
-            PyErr_Format(PyExc_TypeError,
-                         "crypt() argument 1 must be read-only bytes-like "
-                         "object, not %s",
-                         arg->ob_type->tp_name);
+    if (PyBytes_Check(arg)) {
+        input = PyBytes_AS_STRING(arg);
+        size = PyBytes_GET_SIZE(arg);
+    }
+    else if (PyUnicode_Check(arg)) {
+#if PY_MAJOR_VERSION >= 3
+        input = (char *)PyUnicode_AsUTF8AndSize(arg, &size);
+        if (input == NULL) {
+            return NULL;
         }
-        return NULL;
+#else
+        arg = PyUnicode_AsASCIIString(arg);
+        if (arg == NULL) {
+            return NULL;
+        }
+        outputBytes = arc4_ARC4_crypt(self, arg);
+        Py_DECREF(arg);
+        return outputBytes;
+#endif /* PY_MAJOR_VERSION >= 3 */
+    }
+    else {
+        return PyErr_Format(PyExc_TypeError,
+                            "crypt() argument 1 must be read-only bytes-like "
+                            "object, not %s",
+                            Py_TYPE(arg)->tp_name);
     }
     outputBytes = PyBytes_FromStringAndSize(NULL, size);
     output = PyBytes_AS_STRING(outputBytes);
@@ -238,7 +248,8 @@ PyDoc_STRVAR(arc4_ARC4_encrypt_doc,
 static PyMethodDef arc4_ARC4_methods[] = {
     {"decrypt", (PyCFunction)arc4_ARC4_crypt, METH_O, arc4_ARC4_decrypt_doc},
     {"encrypt", (PyCFunction)arc4_ARC4_crypt, METH_O, arc4_ARC4_encrypt_doc},
-    {NULL}};
+    {NULL} /* sentinel */
+};
 
 PyDoc_STRVAR(arc4_ARC4Type_doc,
              "A class represents a session of RC4 stream cipher.\n"
@@ -260,6 +271,8 @@ PyDoc_STRVAR(arc4_ARC4Type_doc,
              "You have to initialize an instance in the beginning of each "
              "operations.\n");
 
+PyDoc_STRVAR(arc4_doc, "ARCFOUR (RC4) implementation in Python/C API.");
+
 static PyTypeObject arc4_ARC4Type = {
     PyVarObject_HEAD_INIT(NULL, 0) /* A trailing comma is included */
     "arc4.ARC4",
@@ -302,8 +315,13 @@ static PyTypeObject arc4_ARC4Type = {
 };
 
 #if PY_MAJOR_VERSION >= 3
-static struct PyModuleDef arc4_module = {PyModuleDef_HEAD_INIT, "arc4", NULL,
-                                         -1, NULL};
+static struct PyModuleDef arc4_module = {
+    PyModuleDef_HEAD_INIT, /* m_base */
+    "arc4",                /* m_name */
+    arc4_doc,              /* m_doc */
+    -1,                    /* m_size */
+    NULL                   /* m_methods */
+};
 
 PyMODINIT_FUNC
 PyInit_arc4(void)
@@ -317,10 +335,17 @@ PyInit_arc4(void)
     if (module == NULL) {
         return NULL;
     }
+    if (PyModule_AddStringConstant(module, "__version__",
+                                   Py_STRINGIFY(ARC4_VERSION)) < 0) {
+        Py_DECREF(module);
+        return NULL;
+    }
     Py_INCREF(&arc4_ARC4Type);
-    PyModule_AddStringConstant(module, "__version__",
-                               Py_STRINGIFY(ARC4_VERSION));
-    PyModule_AddObject(module, "ARC4", (PyObject *)&arc4_ARC4Type);
+    if (PyModule_AddObject(module, "ARC4", (PyObject *)&arc4_ARC4Type) < 0) {
+        Py_DECREF(&arc4_ARC4Type);
+        Py_DECREF(module);
+        return NULL;
+    }
     return module;
 }
 #else
@@ -332,13 +357,20 @@ initarc4(void)
     if (PyType_Ready(&arc4_ARC4Type) < 0) {
         return;
     }
-    module = Py_InitModule("arc4", NULL);
+    module = Py_InitModule3("arc4", NULL, arc4_doc);
     if (module == NULL) {
         return;
     }
+    if (PyModule_AddStringConstant(module, "__version__",
+                                   Py_STRINGIFY(ARC4_VERSION)) < 0) {
+        Py_DECREF(module);
+        return;
+    }
     Py_INCREF(&arc4_ARC4Type);
-    PyModule_AddStringConstant(module, "__version__",
-                               Py_STRINGIFY(ARC4_VERSION));
-    PyModule_AddObject(module, "ARC4", (PyObject *)&arc4_ARC4Type);
+    if (PyModule_AddObject(module, "ARC4", (PyObject *)&arc4_ARC4Type) < 0) {
+        Py_DECREF(&arc4_ARC4Type);
+        Py_DECREF(module);
+        return;
+    }
 }
 #endif /* PY_MAJOR_VERSION >= 3 */
